// AXIS.C
//
// Axis translation for X36/45 driver
//


//#define DBG 1

#include <WDM.H>
#include <filter.h>

void ClearAxis(PAXIS_CAL Axis)
{
  unsigned char i;

  for (i=0; i < 3; i++)
  {
    Axis->floating[i] = 1;
    Axis->value[i] = 0xFFFF;
    Axis->deadzone[i] = 0;
  }
}



unsigned short Calculate(PAXIS_CAL Axis, unsigned short raw, unsigned char bits)
{
    unsigned char calcCenter = 0;
    unsigned long output, opmax;


    // check all points if necessary to update.
    // min
    if ((Axis->floating[0]) && ((raw < Axis->value[0]) || (Axis->value[0] == 0xFFFF)))
    {
      Axis->value[0] = raw;
      calcCenter = 1;
    }
    // max
    if ((Axis->floating[2]) && ((raw > Axis->value[2]) || (Axis->value[2] == 0xFFFF)))
    {
      Axis->value[2] = raw;
      calcCenter = 1;
    }
    // center
    if ((Axis->floating[1]) && ((calcCenter) || (Axis->value[1] == 0xFFFF)))
    {
      Axis->value[1] = ((Axis->value[2] - Axis->value[0])/2) + Axis->value[0];
    }

    // sanity test
    if ((Axis->value[0] == Axis->value[1]) || (Axis->value[1] == Axis->value[2]))
      return 0;

    // compute max value
    calcCenter = (raw >= Axis->value[1]); // above center
    opmax = 0x7FFF >> (16-bits);    
    
    DbgPrn(("Abv: %d   OpMax: %ld   r: %d  c1: %d c2: %d\n",calcCenter,opmax,raw,(Axis->value[calcCenter] + Axis->deadzone[calcCenter]),(Axis->value[calcCenter+1]-Axis->deadzone[calcCenter+1])));
    // do scaling and deadzone test
    if (raw <= (Axis->value[calcCenter] + Axis->deadzone[calcCenter]))
      output = 0;
    else if (raw >= (Axis->value[calcCenter+1] - Axis->deadzone[calcCenter+1]))
      output = opmax;
    else
      output = (((unsigned long)raw - (unsigned long)(Axis->value[calcCenter] + Axis->deadzone[calcCenter])) * opmax) / ((unsigned long)(Axis->value[calcCenter+1]-Axis->deadzone[calcCenter+1]) - (unsigned long)(Axis->value[calcCenter]+Axis->deadzone[calcCenter]));
      
    output += (calcCenter) * (opmax+1);

    return (unsigned short) output;
}



void ClearCalibration(struct DEVICE_EXTENSION  *DeviceExtension)
{
  unsigned char i;

  DbgPrn(("X36C: Clearing calibration data.\n"));
  for (i=0; i<NUM_AXIS; i++)
    ClearAxis(&(DeviceExtension->Calibration[i]));
}

